/*****************************************************************************/
/*                                                                           */
/*				     cbm.h			       	     */
/*                                                                           */
/*			CBM system-specific definitions			     */
/*                                                                           */
/*                                                                           */
/*                                                                           */
/* (c) 1998-2011, Ullrich von Bassewitz                                      */
/*                Roemerstrasse 52                                           */
/*                D-70794 Filderstadt                                        */
/* EMail:         uz@cc65.org                                                */
/*                                                                           */
/*                                                                           */
/* This software is provided 'as-is', without any expressed or implied       */
/* warranty.  In no event will the authors be held liable for any damages    */
/* arising from the use of this software.                                    */
/*                                                                           */
/* Permission is granted to anyone to use this software for any purpose,     */
/* including commercial applications, and to alter it and redistribute it    */
/* freely, subject to the following restrictions:                            */
/*                                                                           */
/* 1. The origin of this software must not be misrepresented; you must not   */
/*    claim that you wrote the original software. If you use this software   */
/*    in a product, an acknowledgment in the product documentation would be  */
/*    appreciated but is not required.                                       */
/* 2. Altered source versions must be plainly marked as such, and must not   */
/*    be misrepresented as being the original software.                      */
/* 3. This notice may not be removed or altered from any source              */
/*    distribution.                                                          */
/*                                                                           */
/*****************************************************************************/



#ifndef _CBM_H
#define _CBM_H



/* Check for errors */
#if !defined(__CBM__)
#  error This module may be used only when compiling for CBM machines!
#endif



/* We need NULL. */

#if !defined(_STDDEF_H)
#  include <stddef.h>
#endif

/* Load the system-specific files here, if needed. */
#if   defined(__C64__)    && !defined(_C64_H)
#  include <c64.h>
#elif defined(__VIC20__)  && !defined(_VIC20_H)
#  include <vic20.h>
#elif defined(__C128__)   && !defined(_C128_H)
#  include <c128.h>
#elif defined(__PLUS4__)  && !defined(_PLUS4_H)
#  include <plus4.h>
#elif defined(__C16__)    && !defined(_C16_H)
#  include <c16.h>
#elif defined(__CBM510__) && !defined(_CBM510_H)
#  include <cbm510.h>
#elif defined(__CBM610__) && !defined(_CBM610_H)
#  include <cbm610.h>
#elif defined(__PET__)    && !defined(_PET_H)
#  include <pet.h>
#endif



/*****************************************************************************/
/*                                 Variables                                 */
/*****************************************************************************/



/* The file stream implementation and the POSIX I/O functions will use the
 * following variables to determine the file type and the disk unit to use.
 */
extern unsigned char _curunit;  /* Defaults to current when program started */
extern char _filetype;          /* Defaults to 'u' */



/*****************************************************************************/
/*                       Character-codes (CBM charset)                       */
/*****************************************************************************/



#ifdef __CC65__
/* ASCII-to-PetSCII character-map translations:
 * These lines convert characters that are not in CBM's character-set
 * into codes with shapes that resemble the ASCII shapes.
 */
#pragma charmap(0x5c,0xbf)	/* '\\' */
#pragma charmap(0x5f,0xa4)	/* '_'  thinner than built-in one */
#pragma charmap(0x60,0xad)	/* '`'  */
#pragma charmap(0x7b,0xb3)	/* '{'  */
#pragma charmap(0x7c,0xdd)	/* '|'  "official" PetSCII code */
#pragma charmap(0x7d,0xab)	/* '}'  */
#pragma charmap(0x7e,0xb1)	/* '~'  */
#endif



#define CH_HLINE	192
#define CH_VLINE	221
#define CH_ULCORNER	176
#define CH_URCORNER	174
#define CH_LLCORNER	173
#define CH_LRCORNER	189
#define CH_TTEE		178
#define CH_BTEE		177
#define CH_LTEE		171
#define CH_RTEE		179
#define CH_CROSS	219
#define CH_CURS_UP	145
#define CH_CURS_DOWN	 17
#define CH_CURS_LEFT	157
#define CH_CURS_RIGHT	 29
#define CH_PI		222
#define CH_HOME		 19
#define CH_DEL		 20
#define CH_INS		148
#define CH_ENTER	 13
#define CH_STOP		  3
#define CH_LIRA		 92
#define CH_ESC		 27



/*****************************************************************************/
/*                Definitions for directory reading functions                */
/*****************************************************************************/



/* CBM FILE TYPES */
#define CBM_T_DEL    0
#define CBM_T_SEQ    1
#define CBM_T_PRG    2
#define CBM_T_USR    3
#define CBM_T_REL    4
#define CBM_T_CBM    5          /* 1581 sub-partition */
#define CBM_T_DIR    6          /* IDE64 and CMD sub-directory */
#define CBM_T_LNK    7          /* IDE64 soft-link */
#define CBM_T_VRP    8          /* Vorpal fast-loadable format */
#define CBM_T_OTHER  9          /* File-type not recognized */
#define CBM_T_HEADER 10         /* Disk header / title */

/* CBM FILE ACCESS */
#define CBM_A_RO    1           /* Read only   */
#define CBM_A_WO    2           /* Write only  */
#define CBM_A_RW    3           /* Read, Write */

struct cbm_dirent {
             char name[17];     /* File name in PetSCII, limited to 16 chars */
    unsigned int  size;         /* Size, in 254-/256-byte blocks */
    unsigned char type;
    unsigned char access;
};



/*****************************************************************************/
/*                               Machine info                                */
/*****************************************************************************/



#define TV_NTSC         0
#define TV_PAL          1
#define TV_OTHER        2

unsigned char __fastcall__ get_tv (void);
/* Return the video mode the machine is using. */



/*****************************************************************************/
/*                           CBM kernal functions                            */
/*****************************************************************************/



/* Constants to use with cbm_open() for openning a file for reading or
 * writing without the need to append ",r" or ",w" to the filename.
 *
 * e.g., cbm_open(2, 8, CBM_READ, "0:data,s");
 */
#define CBM_READ	0	/* default is ",p" */
#define CBM_WRITE	1	/* ditto */
#define CBM_SEQ		2	/* default is ",r" -- or ",s" when writing */

/* Kernal-level functions */
void __fastcall__ cbm_k_setlfs (unsigned char LFN, unsigned char DEV,
                                unsigned char SA);
void __fastcall__ cbm_k_setnam (const char* Name);
unsigned int __fastcall__ cbm_k_load(unsigned char flag, unsigned addr);
unsigned char __fastcall__ cbm_k_save(unsigned int start, unsigned int end);
unsigned char cbm_k_open (void);
void __fastcall__ cbm_k_close (unsigned char FN);
signed char cbm_k_readst (void);
unsigned char __fastcall__ cbm_k_chkin (unsigned char FN);
unsigned char __fastcall__ cbm_k_ckout (unsigned char FN);
unsigned char cbm_k_basin (void);
unsigned char cbm_k_getin (void);
void __fastcall__ cbm_k_bsout (unsigned char C);
void cbm_k_clrch (void);
void cbm_clall (void);



/*****************************************************************************/
/*                       BASIC-like file I/O functions                       */
/*****************************************************************************/



/* The cbm_* I/O functions below set _oserror (see errno.h),
 * in case of an error.
 *
 * error-code	BASIC error
 * ----------	-----------
 *	 1  =	too many files
 *	 2  =	file open
 *	 3  =	file not open
 *	 4  =	file not found
 *	 5  =	device not present
 *	 6  =	not input-file
 *	 7  =	not output-file
 *	 8  =	missing file-name
 *	 9  =	illegal device-number
 *
 *	10  =	STOP-key pushed
 *	11  =	general I/O-error
 */



unsigned int cbm_load (const char* name, unsigned char device, void* data);
/* Loads file "name", from given device, to given address -- or, to the load
 * address of the file if "data" is the null pointer (like load"name",8,1
 * in BASIC).
 * Returns number of bytes that were loaded if loading was successful;
 * otherwise 0, "_oserror" contains an error-code, then (see table above).
 */

unsigned char __fastcall__ cbm_save (const char* name, unsigned char device,
                                     const void* addr, unsigned int size);
/* Saves "size" bytes, starting at "addr", to a file.
 * Returns 0 if saving was successful, otherwise an error-code (see table
 * above).
 */

unsigned char __fastcall__ cbm_open (unsigned char lfn, unsigned char device,
                                     unsigned char sec_addr, const char* name);
/* Opens a file. Works just like the BASIC command.
 * Returns 0 if openning was successful, otherwise an error-code (see table
 * above).
 */

void __fastcall__ cbm_close (unsigned char lfn);
/* Closes a file */

int __fastcall__ cbm_read (unsigned char lfn, void* buffer, unsigned int size);
/* Reads up to "size" bytes from a file into "buffer".
 * Returns the number of actually-read bytes, 0 if there are no bytes left.
 * -1 in case of an error; then, _oserror contains an error-code (see table
 * above).  (Remember:  0 means end-of-file; -1 means error.)
 */

int __fastcall__ cbm_write (unsigned char lfn, const void* buffer,
                            unsigned int size);
/* Writes up to "size" bytes from "buffer" to a file.
 * Returns the number of actually-written bytes, or -1 in case of an error;
 * _oserror contains an error-code, then (see above table).
 */

unsigned char __fastcall__ cbm_opendir (unsigned char lfn, unsigned char device);
/* Opens directory listing.
 * Returns 0 if openning directory was successful;
 * otherwise, an error-code corresponding to cbm_open().
 */

unsigned char __fastcall__ cbm_readdir (unsigned char lfn,
                                        struct cbm_dirent* l_dirent);
/* Reads one directory line into cbm_dirent structure.
 * Returns 0 if reading directory-line was successful.
 * Returns non-zero if reading directory failed, or no more file-names to read.
 * Returns 2 on last line.  Then, l_dirent->size = the number of "blocks free."
 */

void __fastcall__ cbm_closedir (unsigned char lfn);
/* Closes directory by cbm_close(lfn) */



/* End of cbm.h */
#endif


